from typing import cast

from sqlalchemy import ColumnElement, exists
from tqdm import tqdm

from reling.app.exceptions import AlgorithmException
from reling.db import single_session
from reling.db.models import Dialogue, DialogueExchangeTranslation, Language, Text, TextSentenceTranslation
from reling.gpt import GPTClient
from reling.types import DialogueExchangeData, Promise
from .exceptions import TranslationExistsException
from .storage import save_dialogue_translation, save_text_translation
from .translation import translate_dialogue_exchanges, translate_text_sentences

__all__ = [
    'translate_dialogue',
    'translate_text',
]


def is_text_translated(text: Text, language: Language) -> bool:
    """Check if a text has already been translated into a target language."""
    with single_session() as session:
        return session.query(exists().where(
            cast(ColumnElement[bool], TextSentenceTranslation.text_id == text.id),
            cast(ColumnElement[bool], TextSentenceTranslation.language_id == language.id),
        )).scalar()


def is_dialogue_translated(dialogue: Dialogue, language: Language) -> bool:
    """Check if a dialogue has already been translated into a target language."""
    with single_session() as session:
        return session.query(exists().where(
            cast(ColumnElement[bool], DialogueExchangeTranslation.dialogue_id == dialogue.id),
            cast(ColumnElement[bool], DialogueExchangeTranslation.language_id == language.id),
        )).scalar()


def translate_text(gpt: Promise[GPTClient], text: Text, language: Language) -> None:
    """
    Translate a text into another language.

    :raises ValueError: If the text is already in the target language.
    :raises TranslationExistsException: If the text has already been translated into the target language.
    :raises AlgorithmException: If there is an issue with the translation algorithm.
    """
    if language.id == text.language_id:
        raise ValueError(f'The text is already in {language.name}.')
    if is_text_translated(text, language):
        raise TranslationExistsException
    sentences = list(tqdm(
        translate_text_sentences(
            gpt=gpt(),
            sentences=[sentence.sentence for sentence in text.sentences],
            source_language=text.language,
            target_language=language,
        ),
        desc=f'Translating text into {language.name}',
        total=len(text.sentences),
    ))
    if len(sentences) != len(text.sentences):
        raise AlgorithmException(
            'The number of translated sentences does not match the number of original sentences. You can try again.',
        )
    save_text_translation(text, language, sentences)


def translate_dialogue(gpt: Promise[GPTClient], dialogue: Dialogue, language: Language) -> None:
    """
    Translate a dialogue into another language.

    :raises ValueError: If the dialogue is already in the target language.
    :raises TranslationExistsException: If the dialogue has already been translated into the target language.
    :raises AlgorithmException: If there is an issue with the translation algorithm.
    """
    if language.id == dialogue.language_id:
        raise ValueError(f'The dialogue is already in {language.name}.')
    if is_dialogue_translated(dialogue, language):
        raise TranslationExistsException
    exchanges = list(tqdm(
        translate_dialogue_exchanges(
            gpt=gpt(),
            exchanges=[
                DialogueExchangeData(
                    speaker=exchange.speaker,
                    user=exchange.user,
                )
                for exchange in dialogue.exchanges
            ],
            speaker_gender=dialogue.speaker_gender,
            user_gender=dialogue.user_gender,
            source_language=dialogue.language,
            target_language=language,
        ),
        desc=f'Translating dialogue into {language.name}',
        total=len(dialogue.exchanges),
    ))
    if len(exchanges) != len(dialogue.exchanges):
        AlgorithmException(
            'The number of translated exchanges does not match the number of original exchanges. You can try again.',
        )
    save_dialogue_translation(dialogue, language, exchanges)



from .dialogues import Dialogue, DialogueExam, DialogueExamResult, DialogueExchange, DialogueExchangeTranslation
from .languages import Language
from .misc import IdIndex
from .modifiers import Speaker, Style, Topic
from .texts import Text, TextExam, TextExamResult, TextSentence, TextSentenceTranslation

__all__ = [
    'Dialogue',
    'DialogueExam',
    'DialogueExamResult',
    'DialogueExchange',
    'DialogueExchangeTranslation',
    'IdIndex',
    'Language',
    'Speaker',
    'Style',
    'Text',
    'TextExam',
    'TextExamResult',
    'TextSentence',
    'TextSentenceTranslation',
    'Topic',
]



from datetime import datetime, timedelta, UTC

import pytz

__all__ = [
    'format_time',
    'format_time_delta',
    'now',
]

TIME_FORMAT = '%Y-%m-%d, %H:%M'


def format_time(time: datetime) -> str:
    return pytz.utc.localize(time).astimezone().strftime(TIME_FORMAT)


def format_time_delta(delta: timedelta) -> str:
    """Format timedelta as follows: '[X days, ][H:][M]M:SS'."""
    return str(delta).split('.')[0].removeprefix('0:').removeprefix('0')


def now() -> datetime:
    return datetime.now(UTC)



from sqlalchemy.exc import IntegrityError, NoResultFound

from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import CONTENT_ARG, NEW_ID_ARG
from reling.db import single_session
from reling.db.models import IdIndex
from reling.utils.typer import typer_raise

__all__ = [
    'rename',
]


@app.command()
def rename(content: CONTENT_ARG, new_id: NEW_ID_ARG) -> None:
    """Rename a text or dialogue."""
    set_default_content(content)
    old_id = content.id
    with single_session() as session:
        try:
            id_index_item = session.query(IdIndex).filter_by(id=content.id).one()
        except NoResultFound:
            typer_raise(f'There is no content with the ID "{content.id}".')  # Should never happen
        id_index_item.id = new_id
        content.id = new_id
        try:
            session.commit()
            set_default_content(content)
            print(f'Renamed "{old_id}" to "{new_id}".')
        except IntegrityError:
            typer_raise(f'The ID "{new_id}" is already in use.')



from typing import Callable

from reling.types import Promise

__all__ = [
    'name_function',
    'named_function',
    'promisify',
]


def name_function[T: Callable](function: T, name: str) -> T:
    """Set the name of a function."""
    function.__name__ = name
    return function


def named_function[T: Callable](name: str) -> Callable[[T], T]:
    """Decorator to set the name of a function."""
    return lambda function: name_function(function, name)


def promisify[T](value: T) -> Promise[T]:
    """Create a promise that returns a value."""
    return lambda: value



from enum import StrEnum

from reling.db.enums import Gender

__all__ = [
    'Voice',
]


class Voice(StrEnum):
    ALLOY = 'alloy'
    ECHO = 'echo'
    FABLE = 'fable'
    ONYX = 'onyx'
    NOVA = 'nova'
    SHIMMER = 'shimmer'

    @property
    def gender(self) -> Gender:
        match self:
            case Voice.ALLOY:
                return Gender.NONBINARY
            case Voice.ECHO:
                return Gender.MALE
            case Voice.FABLE:
                return Gender.NONBINARY
            case Voice.ONYX:
                return Gender.MALE
            case Voice.NOVA:
                return Gender.FEMALE
            case Voice.SHIMMER:
                return Gender.FEMALE
            case _:
                raise ValueError(f'Unknown voice: {self}')



from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import CONTENT_ARG, FORCE_OPT
from reling.db import single_session
from reling.db.models import IdIndex
from reling.utils.typer import typer_raise

__all__ = [
    'delete',
]


@app.command()
def delete(content: CONTENT_ARG, force: FORCE_OPT = False) -> None:
    """Delete a text or dialogue."""
    set_default_content(content)

    if not force and not input(f'Are you sure you want to delete "{content.id}"? (y/n): ').lower().startswith('y'):
        typer_raise('Operation canceled.')

    with single_session() as session:
        session.delete(content)
        session.query(IdIndex).filter_by(id=content.id).delete()
        session.commit()
    set_default_content(None)
    print(f'Deleted "{content.id}".')



from reling.app.default_content import get_default_content_id
from reling.db import single_session
from reling.db.models import Dialogue, Text
from reling.utils.typer import typer_raise

__all__ = [
    'find_content',
]


def find_content(content_id: str, last_content_marker: str) -> Text | Dialogue | None:
    """Find a text or dialogue by its ID."""
    if content_id == last_content_marker:
        content_id = get_default_content_id()
        if content_id is None:
            typer_raise('No content has been interacted with yet')
    with single_session() as session:
        return session.get(Text, content_id) or session.get(Dialogue, content_id)



from io import BytesIO

from reling.db.models import Language
from reling.types import Speed
from reling.utils.typer import typer_raise
from .tts_client import TTSClient
from .voices import Voice

__all__ = [
    'GTTSClient',
]

FORMAT = 'mp3'


class GTTSClient(TTSClient):
    _language_code: str

    def __init__(self, language: Language) -> None:
        self._language_code = language.short_code

    def read(self, text: str, _voice: Voice, speed: Speed) -> None:
        if not text.strip():
            return

        # Only import the modules if audio is used
        from gtts import gTTS, gTTSError
        from pydub import AudioSegment
        from pydub.playback import play

        try:
            tts = gTTS(
                text=text,
                lang=self._language_code,
                slow=speed == Speed.SLOW,
            )
        except (AssertionError, RuntimeError, ValueError) as e:
            typer_raise(f'Failed to generate audio: {e}')
        audio_data = BytesIO()
        try:
            tts.write_to_fp(audio_data)
            audio_data.seek(0)
            audio_segment = AudioSegment.from_file(audio_data, format=FORMAT)
            play(audio_segment)
        except gTTSError as e:
            typer_raise(f'Failed to generate audio: {e}')
        finally:
            audio_data.close()



from pathlib import Path
from tempfile import TemporaryDirectory

from tqdm import tqdm

from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.exceptions import AlgorithmException
from reling.app.translation import get_dialogue_exchanges, get_text_sentences
from reling.app.types import (
    API_KEY,
    ASR_MODEL,
    CONTENT_ARG,
    LANGUAGE_OPT,
    LANGUAGE_OPT_FROM,
    LISTEN_OPT,
    MODEL,
    READ_LANGUAGE_OPT,
    TTS_MODEL,
)
from reling.asr import ASRClient
from reling.db.models import Dialogue, Language, Text
from reling.gpt import GPTClient
from reling.helpers.pyaudio import ensure_pyaudio
from reling.helpers.voices import pick_voices
from reling.tts import get_tts_client, TTSClient
from reling.utils.functions import promisify
from reling.utils.time import now
from reling.utils.typer import typer_raise
from .input import collect_dialogue_translations, collect_text_translations
from .presentation import present_dialogue_results, present_text_results
from .scoring import score_dialogue_translations, score_text_translations
from .storage import save_dialogue_exam, save_text_exam
from .types import ExchangeWithTranslation, SentenceWithTranslation

__all__ = [
    'exam',
]


@app.command()
def exam(
        api_key: API_KEY,
        model: MODEL,
        tts_model: TTS_MODEL,
        asr_model: ASR_MODEL,
        content: CONTENT_ARG,
        from_: LANGUAGE_OPT_FROM = None,
        to: LANGUAGE_OPT = None,
        read: READ_LANGUAGE_OPT = None,
        listen: LISTEN_OPT = False,
) -> None:
    """
    Test the user's ability to translate content from one language to another.
    If only one language is specified, the content's original language is assumed for the unspecified direction.
    """
    set_default_content(content)
    if read or listen:
        ensure_pyaudio()
    if from_ is None and to is None:
        typer_raise('You must specify at least one language.')
    from_ = from_ or content.language
    to = to or content.language
    if from_ == to:
        typer_raise('The source and target languages are the same.')

    read = read or []
    for language in read:
        if language not in [from_, to]:
            typer_raise(f'Cannot read in {language.name} as it is not the source or target language.')

    (perform_text_exam if isinstance(content, Text) else perform_dialogue_exam)(
        GPTClient(api_key=api_key.get(), model=model.get()),
        content,
        from_,
        to,
        source_tts=(get_tts_client(model=tts_model.get(), api_key=api_key.promise(), language=from_)
                    if from_ in read else None),
        target_tts=(get_tts_client(model=tts_model.get(), api_key=api_key.promise(), language=to)
                    if to in read else None),
        asr=ASRClient(api_key=api_key.get(), model=asr_model.get()) if listen else None,
    )


def perform_text_exam(
        gpt: GPTClient,
        text: Text,
        source_language: Language,
        target_language: Language,
        source_tts: TTSClient | None,
        target_tts: TTSClient | None,
        asr: ASRClient | None,
) -> None:
    """
    Translate the text as needed, collect user translations, score them, save and present the results to the user,
    optionally reading the source and/or target language out loud.
    """
    with TemporaryDirectory() as file_storage:
        source_voice, target_voice = pick_voices(None, None)
        voice_source_tts = source_tts.with_voice(source_voice) if source_tts else None
        voice_target_tts = target_tts.with_voice(target_voice) if target_tts else None

        sentences = get_text_sentences(promisify(gpt), text, source_language)
        original_translations = get_text_sentences(promisify(gpt), text, target_language)

        started_at = now()
        translated = list(collect_text_translations(
            sentences=sentences,
            target_language=target_language,
            source_tts=voice_source_tts,
            asr=asr,
            storage=Path(file_storage),
        ))
        finished_at = now()

        try:
            results = list(tqdm(
                score_text_translations(
                    gpt=gpt,
                    sentences=translated,
                    original_translations=original_translations,
                    source_language=source_language,
                    target_language=target_language,
                ),
                desc='Scoring translations',
                total=len(translated),
            ))
        except AlgorithmException as e:
            typer_raise(e.msg)

        save_text_exam(
            text=text,
            source_language=source_language,
            target_language=target_language,
            read_source=source_tts is not None,
            read_target=target_tts is not None,
            listened=asr is not None,
            started_at=started_at,
            finished_at=finished_at,
            sentences=translated,
            results=results,
        )

        present_text_results(
            sentences=translated,
            original_translations=original_translations,
            results=results,
            duration=finished_at - started_at,
            source_tts=voice_source_tts,
            target_tts=voice_target_tts,
        )


def perform_dialogue_exam(
        gpt: GPTClient,
        dialogue: Dialogue,
        source_language: Language,
        target_language: Language,
        source_tts: TTSClient | None,
        target_tts: TTSClient | None,
        asr: ASRClient | None,
) -> None:
    """
    Translate the dialogue as needed, collect user translations, score them, save and present the results to the user,
    optionally reading the source and/or target language out loud.
    """
    with TemporaryDirectory() as file_storage:
        speaker_voice, user_voice = pick_voices(dialogue.speaker_gender, dialogue.user_gender)
        source_user_tts = source_tts.with_voice(user_voice) if source_tts else None
        target_user_tts = target_tts.with_voice(user_voice) if target_tts else None
        target_speaker_tts = target_tts.with_voice(speaker_voice) if target_tts else None

        exchanges = get_dialogue_exchanges(promisify(gpt), dialogue, source_language)
        original_translations = get_dialogue_exchanges(promisify(gpt), dialogue, target_language)

        started_at = now()
        translated = list(collect_dialogue_translations(
            exchanges=exchanges,
            original_translations=original_translations,
            target_language=target_language,
            source_user_tts=source_user_tts,
            target_speaker_tts=target_speaker_tts,
            asr=asr,
            storage=Path(file_storage),
        ))
        finished_at = now()

        try:
            results = list(tqdm(
                score_dialogue_translations(
                    gpt=gpt,
                    exchanges=translated,
                    original_translations=original_translations,
                    source_language=source_language,
                    target_language=target_language,
                ),
                desc='Scoring translations',
                total=len(translated),
            ))
        except AlgorithmException as e:
            typer_raise(e.msg)

        save_dialogue_exam(
            dialogue=dialogue,
            source_language=source_language,
            target_language=target_language,
            read_source=source_tts is not None,
            read_target=target_tts is not None,
            listened=asr is not None,
            started_at=started_at,
            finished_at=finished_at,
            exchanges=translated,
            results=results,
        )

        present_dialogue_results(
            exchanges=translated,
            original_translations=original_translations,
            results=results,
            duration=finished_at - started_at,
            source_user_tts=source_user_tts,
            target_speaker_tts=target_speaker_tts,
            target_user_tts=target_user_tts,
        )
